import datetime
import logging

t = True
f = False

## =============================================================================

## Release settings in (). Enforced by Makefile using good old sed :-)

c = 'cli'
h = 'html'
m = 'md'

RECORD_AST = f  ## (f)
OUTPUT = h  ## set html as default output (h)
SHOW_OUTPUT = t  ## f is only ever used when testing pre-display (t)
INCLUDE_LINTING = t  ## f when running unit tests to massively speed them up (otherwise every snippet in tests is linted each time) (t)
LOG_LEVEL = logging.INFO  ## (logging.INFO)
## =============================================================================

## When testing user-supplied snippets watch out for the BOM MS inserts via Notepad. AST chokes on it.
## All snippets here should be raw strings (see https://stackoverflow.com/questions/53636723/python-parsing-code-with-new-line-character-in-them-using-ast)
TEST_SNIPPET = r"""
msgs = []
for i in range(20):
    msgs.append(f"i is {i}")
name = 'Grant'
"""

PY3_6 = '3.6'
PY3_7 = '3.7'
PY3_8 = '3.8'

LINUX = 'linux'
WINDOWS = 'windows'
MAC = 'mac'

AST_OUTPUT_XML_FNAME = 'ast_output.xml'

PYTHON_CODE_START = '__python_code_start__'
PYTHON_CODE_END = '__python_code_end__'

MD_PYTHON_CODE_START = '::python'

BRIEF = 'Brief'  ## no spaces; used as labels and as parts of class names in CSS
MAIN = 'Main'
EXTRA = 'Extra'
DETAIL_LEVELS = [BRIEF, MAIN, EXTRA]

ANON_NAME = 'Anonymous'

INT_TYPE = 'int'
FLOAT_TYPE = 'float'
STR_TYPE = 'str'
DATETIME_TYPE = 'datetime'
BOOLEAN_TYPE = 'bool'
LIST_TYPE = 'list'
DICT_TYPE = 'dict'
TUPLE_TYPE = 'tuple'

TYPE2NAME = {
    INT_TYPE: 'integer',
    FLOAT_TYPE: 'float',
    STR_TYPE: 'string',
    DATETIME_TYPE: 'datetime object',
    BOOLEAN_TYPE: 'boolean',
    LIST_TYPE: 'list',
    DICT_TYPE: 'dict',
    TUPLE_TYPE: 'tuple',
}

EXAMPLES_OF_TYPES = {  ## best to include at least three so we have enough to append one and from what is left extend multiple
    INT_TYPE: [123, 9, 17, 20, 100, 2020, 16],
    FLOAT_TYPE: [1.2345, 0.667, 0.1, 0.001, 10.0],
    STR_TYPE: ['apple', 'banana', 'kiwifruit', 'Auckland, New Zealand'],
    DATETIME_TYPE: [
        datetime.date(2020, 4, 4),
        datetime.date(1066, 10, 14),
        datetime.datetime.today(), ],
    BOOLEAN_TYPE: [True, False],
    LIST_TYPE: [[10, 2], [-3, 20], [44, -180]],
    DICT_TYPE: [{'x': 10, 'y': 2}, {'x': -3, 'y': 20}, {'x': 44, 'y': -180}],
    TUPLE_TYPE: [(10, 2), (-3, 20), (44, -180)],
}

STD_NAME = 'std_name'
DICT_KEY_NAME = 'dict_key_name'
OBJ_ATTR_NAME = 'obj_attr_name'

NON_STD_EL_KEYS = ('lineno', 'col_offset')

MAX_BRIEF_FUNC_LOC = 35
MAX_BRIEF_FUNC_ARGS = 6
MIN_BRIEF_DOCSTRING = 3
MIN_BRIEF_NAME = 3
MAX_BRIEF_NESTED_BLOCK = 20
MIN4ANY_OR_ALL = 3
MAX_ITEMS_EVALUATED = 25
MAX_PROJECT_MODULES = 50
MAX_FILE_PATH_IN_HEADING = 75

FUNCTION_LBL = 'function'
METHOD_LBL = 'method'

EMAIL2USE = 'superhelp@p-s.co.nz'

WARNINGS_ONLY_MSG = ("Only displaying warnings. "
    "To see all help, set warnings only option to False")
ALL_HELP_SHOWING_MSG = ("Displaying all help. "
    "To only see warnings, set warnings only option to True")

NO_ADVICE_MESSAGE = ("No advice to give - looks fine :-). But if you think "
    f"there should have been some advice given, contact {EMAIL2USE} "
    "with the subject line 'Advice' and explain. Please include a snippet to "
    "test as well.")
MISSING_ADVICE_MESSAGE = ("If there was some advice you think should have "
    f"been given that wasn't, contact {EMAIL2USE} with the subject line "
    "'Advice' and explain. Please include a snippet to test as well.")
SYSTEM_MESSAGE = 'System message'

## input types
BLOCKS_DETS = 'blocks_dets'
SNIPPET_STR = 'snippet_str'

XKCD_WARNING_WORDS = ['supervolcano', 'seagull', 'garbage disposal']

VERBOSE_FLAG = 'VERBOSE'
INLINE_RE_VERBOSE_FLAG = '(?x)'

SNIPPET_FNAME = 'snippet.py'
LINT_MSG_TYPE = 'msg_type'
LINT_MSG = 'msg'
LINT_LINE_NO = 'line_no'

LINE_FEED = '&#10;'

FORCE_SPLIT = '__force_split__'

## scraped from https://docs.python.org/3/py-modindex.html 2020-04-02
STD_LIBS = ['__future__', '__main__', '_dummy_thread', '_thread', 'aifc',
'argparse', 'array', 'ast', 'asynchat', 'asyncio', 'asyncore', 'atexit',
'audioop', 'base64', 'bdb', 'binascii', 'binhex', 'bisect', 'builtins', 'bz2',
'calendar', 'cgi', 'cgitb', 'chunk', 'cmath', 'cmd', 'code', 'codecs', 'codeop',
'collections', 'colorsys', 'compileall', 'concurrent', 'configparser',
'contextlib', 'contextvars', 'copy', 'copyreg', 'cProfile', 'crypt', 'csv',
'ctypes', 'curses', 'dataclasses', 'datetime', 'dbm', 'decimal', 'difflib',
'dis', 'distutils', 'doctest', 'dummy_threading', 'email', 'encodings',
'ensurepip', 'enum', 'errno', 'faulthandler', 'fcntl', 'filecmp', 'fileinput',
'fnmatch', 'formatter', 'fractions', 'ftplib', 'functools', 'gc', 'getopt',
'getpass', 'gettext', 'glob', 'grp', 'gzip', 'hashlib', 'heapq', 'hmac', 'html',
'http', 'imaplib', 'imghdr', 'imp', 'importlib', 'inspect', 'io', 'ipaddress',
'itertools', 'json', 'keyword', 'lib2to3', 'linecache', 'locale', 'logging',
'lzma', 'mailbox', 'mailcap', 'marshal', 'math', 'mimetypes', 'mmap',
'modulefinder', 'msilib', 'msvcrt', 'multiprocessing', 'netrc', 'nis',
'nntplib', 'numbers', 'operator', 'optparse', 'os', 'ossaudiodev', 'parser',
'pathlib', 'pdb', 'pickle', 'pickletools', 'pipes', 'pkgutil', 'platform',
'plistlib', 'poplib', 'posix', 'pprint', 'profile', 'pstats', 'pty', 'pwd',
'py_compile', 'pyclbr', 'pydoc', 'queue', 'quopri', 'random', 're', 'readline',
'reprlib', 'resource', 'rlcompleter', 'runpy', 'sched', 'secrets', 'select',
'selectors', 'shelve', 'shlex', 'shutil', 'signal', 'site', 'smtpd', 'smtplib',
'sndhdr', 'socket', 'socketserver', 'spwd', 'sqlite3', 'ssl', 'stat',
'statistics', 'string', 'stringprep', 'struct', 'subprocess', 'sunau', 'symbol',
'symtable', 'sys', 'sysconfig', 'syslog', 'tabnanny', 'tarfile', 'telnetlib',
'tempfile', 'termios', 'test', 'textwrap', 'threading', 'time', 'timeit',
'tkinter', 'token', 'tokenize', 'trace', 'traceback', 'tracemalloc', 'tty',
'turtle', 'turtledemo', 'types', 'typing', 'unicodedata', 'unittest', 'urllib',
'uu', 'uuid', 'venv', 'warnings', 'wave', 'weakref', 'webbrowser', 'winreg',
'winsound', 'wsgiref', 'xdrlib', 'xml', 'xmlrpc', 'zipapp', 'zipfile',
'zipimport', 'zlib']

SUPERHELP_PROJECT_OUTPUT = 'superhelp_project_output'
SUPERHELP_GEN_OUTPUT = 'superhelp_output'

HTML_HEAD = f"""\
<head>
<meta charset="utf-8">
<meta content="IE=edge" http-equiv="X-UA-Compatible">
<title>SuperHELP - Help for Humans!</title>
<style type="text/css">
%(internal_css)s
</style>
</head>"""

INTERNAL_CSS = """\
body {
  background-color: white;
  %(margin_css)s
  %(max_width_css)s
}
h1, h2 {
  color: #0072aa;
  font-weight: bold;
}
h1 {
  font-size: 16px;
}
h2 {
  font-size: 14px;
  margin-top: 24px;
}
h3 {
  font-size: 12px;
}
h4 {
  font-size: 11px;
}
h5 {
  font-size: 9px;
  font-style: italic;
}
p {
  font-size: 10px;
}
li {
  font-size: 10px;
}
label {
  font-size: 12px;
}
blockquote {
  font-style: italic;
  color: #666666;
}
blockquote p {
  font-size: 13px;
}
svg {
  height: 50px;
  width: 60px;
}
.warning {
  border-radius: 2px;
  padding: 12px 6px 6px 12px;
  margin: 10px 0 0 0;
  border: 2px solid #0072aa;
}

#star {
  padding: 9px 10px 10px 10px;
  background-color: #0072aa;
  color: white;
  font-size: 15px;
  font-weight: bold;
  border-radius: 8px;
  text-align: center;
}
#star a:link, #star a:visited, #star a:hover, #star a:active {
  color: white;
  border-bottom: 1px solid white;
  text-decoration: none;
}
#star #grow {
  font-size: 20px;
}

.help {
  display: none;
}
.help.help-visible {
  display: inherit;
}
%(code_css)s
"""

LOGO_SVG = """\
<svg
   xmlns:dc="http://purl.org/dc/elements/1.1/"
   xmlns:cc="http://creativecommons.org/ns#"
   xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#"
   xmlns:svg="http://www.w3.org/2000/svg"
   xmlns="http://www.w3.org/2000/svg"
   xmlns:sodipodi="http://sodipodi.sourceforge.net/DTD/sodipodi-0.dtd"
   xmlns:inkscape="http://www.inkscape.org/namespaces/inkscape"
   width="100.11089mm"
   height="79.607109mm"
   viewBox="0 0 100.11089 79.607109"
   version="1.1"
   id="svg8"
   inkscape:version="0.92.4 (5da689c313, 2019-01-14)"
   sodipodi:docname="superhelp_logo.svg"
   inkscape:export-filename="/home/g/projects/superhelp/superhelp/store/superhelp_logo.png"
   inkscape:export-xdpi="96"
   inkscape:export-ydpi="96">
  <defs
     id="defs2">
    <inkscape:path-effect
       effect="skeletal"
       id="path-effect4972"
       is_visible="true"
       pattern="m 60.80692,207.585 -0.62276,-0.0405 -0.15389,-0.6048 0.527651,-0.33326 0.479996,0.39885 z"
       copytype="repeated"
       prop_scale="1"
       scale_y_rel="false"
       spacing="0"
       normal_offset="0"
       tang_offset="0"
       prop_units="false"
       vertical_pattern="false"
       fuse_tolerance="0" />
    <inkscape:path-effect
       effect="skeletal"
       id="path-effect4944"
       is_visible="true"
       pattern="m 58.866112,205.1261 h 0.701582 v 0.66817 h -0.701582 z"
       copytype="single_stretched"
       prop_scale="1"
       scale_y_rel="false"
       spacing="0"
       normal_offset="0"
       tang_offset="0"
       prop_units="false"
       vertical_pattern="false"
       fuse_tolerance="0" />
    <inkscape:path-effect
       effect="skeletal"
       id="path-effect4907"
       is_visible="true"
       pattern="m 61.427734,769.66211 v 9.31641 h 7.748047 v -9.31641 z m 8.271485,0.0156 v 9.31446 h 7.748047 v -9.31446 z"
       copytype="repeated"
       prop_scale="1"
       scale_y_rel="false"
       spacing="0"
       normal_offset="0"
       tang_offset="0"
       prop_units="false"
       vertical_pattern="false"
       fuse_tolerance="0" />
    <inkscape:path-effect
       effect="skeletal"
       id="path-effect4889"
       is_visible="true"
       pattern="m 61.427734,769.66211 v 9.31641 h 7.748047 v -9.31641 z m 8.271485,0.0156 v 9.31446 h 7.748047 v -9.31446 z"
       copytype="repeated_stretched"
       prop_scale="1"
       scale_y_rel="false"
       spacing="0"
       normal_offset="0"
       tang_offset="0"
       prop_units="false"
       vertical_pattern="false"
       fuse_tolerance="0" />
    <inkscape:path-effect
       effect="skeletal"
       id="path-effect4861"
       is_visible="true"
       pattern="M 0,0 H 1"
       copytype="single_stretched"
       prop_scale="1"
       scale_y_rel="false"
       spacing="0"
       normal_offset="0"
       tang_offset="0"
       prop_units="false"
       vertical_pattern="false"
       fuse_tolerance="0" />
    <inkscape:path-effect
       effect="skeletal"
       id="path-effect4820"
       is_visible="true"
       pattern="m 61.427734,769.66211 v 9.31641 h 7.748047 v -9.31641 z m 8.271485,0.0156 v 9.31446 h 7.748047 v -9.31446 z"
       copytype="repeated_stretched"
       prop_scale="1"
       scale_y_rel="false"
       spacing="0"
       normal_offset="0"
       tang_offset="0"
       prop_units="false"
       vertical_pattern="false"
       fuse_tolerance="0" />
    <inkscape:path-effect
       effect="skeletal"
       id="path-effect4721"
       is_visible="true"
       pattern="m 61.427734,769.66211 v 9.31641 h 7.748047 v -9.31641 z m 8.271485,0.0156 v 9.31446 h 7.748047 v -9.31446 z"
       copytype="repeated"
       prop_scale="1"
       scale_y_rel="false"
       spacing="0.1"
       normal_offset="0"
       tang_offset="0"
       prop_units="false"
       vertical_pattern="false"
       fuse_tolerance="0" />
    <inkscape:path-effect
       effect="skeletal"
       id="path-effect4694"
       is_visible="true"
       pattern="M 0,0 H 1"
       copytype="single_stretched"
       prop_scale="1"
       scale_y_rel="false"
       spacing="0"
       normal_offset="0"
       tang_offset="0"
       prop_units="false"
       vertical_pattern="false"
       fuse_tolerance="0" />
    <inkscape:path-effect
       effect="skeletal"
       id="path-effect935"
       is_visible="true"
       pattern="M 0,0 H 1"
       copytype="single_stretched"
       prop_scale="1"
       scale_y_rel="false"
       spacing="0"
       normal_offset="0"
       tang_offset="0"
       prop_units="false"
       vertical_pattern="false"
       fuse_tolerance="0" />
  </defs>
  <sodipodi:namedview
     id="base"
     pagecolor="#ffffff"
     bordercolor="#666666"
     borderopacity="1.0"
     inkscape:pageopacity="1"
     inkscape:pageshadow="2"
     inkscape:zoom="1.979899"
     inkscape:cx="76.737316"
     inkscape:cy="121.00951"
     inkscape:document-units="mm"
     inkscape:current-layer="layer1"
     showgrid="false"
     inkscape:window-width="1869"
     inkscape:window-height="1056"
     inkscape:window-x="1971"
     inkscape:window-y="24"
     inkscape:window-maximized="1"
     fit-margin-top="0"
     fit-margin-left="0"
     fit-margin-right="0"
     fit-margin-bottom="0"
     inkscape:pagecheckerboard="false" />
  <metadata
     id="metadata5">
    <rdf:RDF>
      <cc:Work
         rdf:about="">
        <dc:format>image/svg+xml</dc:format>
        <dc:type
           rdf:resource="http://purl.org/dc/dcmitype/StillImage" />
        <dc:title />
      </cc:Work>
    </rdf:RDF>
  </metadata>
  <g
     inkscape:label="Layer 1"
     inkscape:groupmode="layer"
     id="layer1"
     transform="translate(-53.753912,-82.960833)">
    <g
       transform="translate(-57.326112,-175.59945)"
       id="g9171">
      <path
         inkscape:connector-curvature="0"
         id="path4991-3-6"
         transform="matrix(0.26458333,0,0,0.26458333,-8.3176424,-7.9183424)"
         d="m 591.57422,1011.6934 a 83.098512,82.305791 18.437406 0 0 -53.05664,20.1914 83.098512,82.305791 18.437406 0 0 -26.49805,78.3183 83.098512,82.305791 18.437406 0 0 54.18164,61.5059 l -0.0371,0.1113 123.08203,41.0332 -2.51171,7.6485 -123.47461,-41.1661 v 0 l -77.85938,-25.957 a 82.244946,81.781039 18.437412 0 0 -1.26562,38.1523 82.244946,81.781039 18.437412 0 0 54.88671,61.3047 l 0.004,-0.01 123.56055,41.1914 -0.008,0.021 a 82.216504,82.216504 0 0 0 0.46094,0.1309 l 0.33594,0.1113 0.006,-0.016 a 82.216504,82.216504 0 0 0 80.125,-16.7149 82.216504,82.216504 0 0 0 25.77344,-78.4648 82.216504,82.216504 0 0 0 -53.91602,-60.75 v 0 l -0.16992,-0.057 a 82.216504,82.216504 0 0 0 -1.30274,-0.4551 l -0.006,0.02 -121.01758,-40.3458 2.64843,-7.83 119.71485,39.9121 -0.002,0.01 78.07226,26.0293 c 3.44134,-12.4392 14.87561,-39.3166 12.27344,-51.9727 -5.95174,-28.611 -37.64693,-38.2382 -65.52734,-47.4199 -0.10692,-0.035 -0.21742,-0.072 -0.32422,-0.1074 l -121.63282,-40.5508 -0.0117,0.033 a 83.098512,82.305791 18.437406 0 0 -26.5039,-3.9062 z"
         style="fill:#0072aa;fill-opacity:1;stroke:none;stroke-width:32.70064163;stroke-miterlimit:4;stroke-dasharray:none;stroke-opacity:1" />
      <path
         transform="rotate(18.437412)"
         sodipodi:open="true"
         d="m 273.07885,208.13282 a 3.2002347,3.2002347 0 0 1 3.20327,-3.19481 3.2002347,3.2002347 0 0 1 3.1972,3.20089 3.2002347,3.2002347 0 0 1 -3.1985,3.19958 3.2002347,3.2002347 0 0 1 -3.20197,-3.19611"
         sodipodi:end="3.1403035"
         sodipodi:start="3.1432891"
         sodipodi:ry="3.2002347"
         sodipodi:rx="3.2002347"
         sodipodi:cy="208.13824"
         sodipodi:cx="276.27908"
         sodipodi:type="arc"
         id="path5097-9-2"
         style="fill:#ffffff;fill-opacity:1;stroke:none;stroke-width:6.38603258;stroke-miterlimit:4;stroke-dasharray:none;stroke-opacity:1" />
      <path
         transform="rotate(18.437412)"
         sodipodi:open="true"
         d="m 227.2124,219.8216 a 3.2002347,3.2002347 0 0 1 3.20327,-3.19481 3.2002347,3.2002347 0 0 1 3.19719,3.20089 3.2002347,3.2002347 0 0 1 -3.19849,3.19958 3.2002347,3.2002347 0 0 1 -3.20197,-3.19611"
         sodipodi:end="3.1403035"
         sodipodi:start="3.1432891"
         sodipodi:ry="3.2002347"
         sodipodi:rx="3.2002347"
         sodipodi:cy="219.82703"
         sodipodi:cx="230.41263"
         sodipodi:type="arc"
         id="path5097-0-7-8"
         style="fill:#ffffff;fill-opacity:1;stroke:none;stroke-width:6.38603258;stroke-miterlimit:4;stroke-dasharray:none;stroke-opacity:1" />
      <path
         transform="rotate(18.437412)"
         sodipodi:open="true"
         d="m 258.29068,242.97 a 3.2002347,3.2002347 0 0 1 3.20327,-3.1948 3.2002347,3.2002347 0 0 1 3.19719,3.20089 3.2002347,3.2002347 0 0 1 -3.1985,3.19958 3.2002347,3.2002347 0 0 1 -3.20197,-3.19611"
         sodipodi:end="3.1403035"
         sodipodi:start="3.1432891"
         sodipodi:ry="3.2002347"
         sodipodi:rx="3.2002347"
         sodipodi:cy="242.97543"
         sodipodi:cx="261.49091"
         sodipodi:type="arc"
         id="path5097-0-9-7-8"
         style="fill:#ffffff;fill-opacity:1;stroke:none;stroke-width:6.38603258;stroke-miterlimit:4;stroke-dasharray:none;stroke-opacity:1" />
    </g>
  </g>
</svg>
"""



"""
Add help modules inside this folder.

To add more helpers, just declare more helper functions inside the helpers
modules with the @..._help decorators :-). Make sure the first paragraph of the
docstring is a good, user-facing description of its purpose so it can be
automatically processed into lists of available help in SuperHELP.

So which help decorator to use?

Is the advice for the snippet as a whole? If so, are you processing the raw
snippet string or the XML code block elements? If processing the snippet string
e.g. passing into flake8 linter use snippet_str_help; if processing XML use
all_blocks_help.

If looking at individual code blocks, are you prefiltering or looking at every
sort? If prefiltering use filt_block_help; if not use any_block_help.
Simple really :-)

The basic pattern within an help function is:

* Get the correct elements and see if the target pattern is found e.g. a value
  being assigned to a name

* If not, exit returning None

* Otherwise create all the different message parts ready to assemble in the
  message.

  Some parts will have two versions - one for the first time the message appears
  for a block of the code snippet; and one for subsequent appearances. These
  "repeat" versions are sometimes empty strings; other times they are simply
  much shorter versions.

  Don't manually try to dedent etc the message parts - use the layout_comment
  function and follow the example of other helper modules. It is very easy to
  break markdown in ways which mess up the terminal output.

* Assemble the message. There can be up to three parts: brief, main, and extra.
  Only the brief component is mandatory. If not supplied, the main component is
  just a repeat of the brief component.

* Add to the appropriate test module. Only test helpers within the module.
"""
import builtins
from collections import namedtuple
from importlib import import_module
import keyword
from pkgutil import iter_modules
import sys
from textwrap import dedent

from .. import conf
from ..gen_utils import get_docstring_start, layout_comment as layout


FILT_BLOCK_HELPERS = []  ## block-based helpers which only apply to blocks filtered to contain specified element types

FiltHelperDets = namedtuple('FilteredHelperDets',
    'helper_name, helper, xpath, warning')
FiltHelperDets.__doc__ += ('\n\nDetails for block-based helpers that only '
    'apply to blocks filtered to contain specified elements')
FiltHelperDets.helper.__doc__ = ('Functions which takes prefiltered '
    'block dets containing the required elements and return message')
FiltHelperDets.xpath.__doc__ = ('xpath filtering to get specified elements '
    'e.g. body/Assign/value/Str')

ANY_BLOCK_HELPERS = []  ## block-based helpers which apply to all blocks

AnyBlockHelperDets = namedtuple('AnyBlockHelperDets',
    'helper_name, helper, warning')
AnyBlockHelperDets.__doc__ += (
    '\n\nDetails for block-based helpers that work on each block')
AnyBlockHelperDets.helper.__doc__ = ('Functions which take block dets '
    '(including element and block code string) and return message')

ALL_BLOCKS_HELPERS = []  ## all-blocks-based helpers (have to look at multiple blocks together at once)
AllBlocksHelperDets = namedtuple('AllBlocksHelperDets',
    'helper_name, helper, warning, input_type')
AllBlocksHelperDets.__doc__ += (
    '\n\nDetails for helpers that work on all blocks together')
AllBlocksHelperDets.helper.__doc__ = ('Functions which take blocks dets '
    '(multiple) and return message')
AllBlocksHelperDets.input_type.__doc__ = (
    'Input type for all blocks helper functions')

SNIPPET_STR_HELPERS = []  ## works on entire code snippet as a single string

SnippetStrHelperDets = namedtuple('SnippetStrHelperDets',
    'helper_name, helper, warning, input_type')
SnippetStrHelperDets.__doc__ += (
    '\n\nDetails for helpers that work on code snippet as a single string')
SnippetStrHelperDets.helper.__doc__ = ('Functions which take the snippet as a'
    ' single code str and return message')
SnippetStrHelperDets.input_type.__doc__ = (
    'Input type for all blocks helper functions')

def filt_block_help(*, xpath, warning=False):
    """
    Simple decorator that registers an unchanged helper function in the list of
    FILT_BLOCK_HELPERS.

    :param str xpath: Used by xpath on the block element being examined. Can
     only use XPath 1.0 syntax.
    :param bool warning: tags messages as warning or not - up to displayer
     e.g. HTML to decide what to do with that information, if anything.
    """
    def decorator(func):
        """
        :param func func: func expecting block_dets
        """
        FILT_BLOCK_HELPERS.append(
            FiltHelperDets(
                f"{func.__module__}.{func.__name__}", func, xpath, warning))
        return func
    return decorator

def any_block_help(*, warning=False):
    """
    Simple decorator that registers an unchanged helper function in the list of
    ANY_BLOCK_HELPERS.

    :param bool warning: tags messages as warning or not - up to displayer
     e.g. HTML to decide what to do with that information, if anything.
    """
    def decorator(func):
        """
        :param func func: func expecting block_dets
        """
        ANY_BLOCK_HELPERS.append(
            AnyBlockHelperDets(
                f"{func.__module__}.{func.__name__}", func, warning))
        return func
    return decorator

def all_blocks_help(*, warning=False):
    """
    Use when processing XML for all blocks at once. Probably looking for first
    instance of something rather than processing code block by code block.

    Simple decorator that registers an unchanged helper function in the list of
    ALL_BLOCKS_HELPERS.

    :param bool warning: tags messages as warning or not - up to displayer
     e.g. HTML to decide what to do with that information, if anything.
    """
    def decorator(func):
        """
        :param func func: func expecting blocks_dets (note plural)
        """
        ALL_BLOCKS_HELPERS.append(
            AllBlocksHelperDets(f"{func.__module__}.{func.__name__}", func,
                warning, conf.BLOCKS_DETS))
        return func
    return decorator

def snippet_str_help(*, warning=False):
    """
    Use when processing the snippet string e.g. passing into flake8 linter.

    Simple decorator that registers an unchanged helper function in the list of
    SNIPPET_STR_HELPERS.

    :param bool warning: tags messages as warning or not - up to displayer
     e.g. HTML to decide what to do with that information, if anything.
    """
    def decorator(func):
        """
        :param func func: func expecting a single code string for the entire
         snippet as input
        """
        SNIPPET_STR_HELPERS.append(
            SnippetStrHelperDets(f"{func.__module__}.{func.__name__}", func,\
                warning, conf.SNIPPET_STR))
        return func
    return decorator

def load_helpers():
    this_module = sys.modules[__name__]
    submodules = iter_modules(
        this_module.__path__,
        this_module.__name__ + '.')
    for submodule in submodules:
        import_module(submodule.name)

def get_helper_comments():
    helper_comments = []
    all_helpers_dets = (
        FILT_BLOCK_HELPERS + ANY_BLOCK_HELPERS + ALL_BLOCKS_HELPERS)
    all_helpers_dets.sort(key=lambda helper_dets: helper_dets.helper.__module__)
    for helper_dets in all_helpers_dets:
        docstring = helper_dets.helper.__doc__
        helper_comment = get_docstring_start(docstring)
        source = helper_dets.helper.__module__.split('.')[-1]
        warning = 'Warning: ' if helper_dets.warning else ''
        helper_comments.append((helper_comment, source, warning))
    return helper_comments

## =============================================================================

## Put into functions so layout etc overheads only incurred in cases when
## actually used

def is_reserved_name(name):
    is_reserved = name in set(keyword.kwlist + dir(builtins) + conf.STD_LIBS)
    return is_reserved

def get_aop_msg():
    return layout("""

        Some aspects of code apply to lots of different parts of the code base
        e.g. logging, or security. These can be thought of as "cross-cutting
        concerns" from an Aspect-Oriented Programming point of view. The problem
        is that a lot of code gets repeated e.g. if every function has to
        implement its own logging or has to handle its own security. Even if
        there is some shared functionality it can often require special wiring
        into other code. In short, coping with cross-cutting concerns can be
        painful.

        Decorators and Context Managers are ways Python provides of
        concentrating cross-cutting concerns into single pieces of code in an
        elegant way. They DRY out your code (see
        <https://en.wikipedia.org/wiki/Don%27t_repeat_yourself>). And because
        they are very widely used, it is worth learning how to use them, even if
        you don't take the step of writing your own.

    """)

def get_unpacking_msg():
    return (
        layout("""\
            Unpacking is much more pythonic than using indexes to pull a
            sequence apart into names (variables). For example:

            """)
        +
        layout("""\

            #### Un-pythonic :-(

            location = (-37, 174, 'Auckland', 'Mt Albert')
            lat = location[0]
            lon = location[1]
            city = location[2]
            suburb = location[3]

            #### Pythonic :-)

            lat, lon, city, suburb = location
            """, is_code=True)
        +
        layout(f"""\

            If you don't need all the values you can indicate which you want to
            ignore or even mop up multiple unused values into a single value
            using an asterisk.

            For example:

            """)
        +
        layout("""\
            lat, lon, _city, _suburb = location
            """, is_code=True)
        +
        layout(f"""\

            or:

            """)
        +
        layout(f"""\
            lat, lon, *_ = location
            """, is_code=True)
        +
        layout("""\

            or:

            """)
        +
        layout("""\
            lat, lon, *unused = location
            """, is_code=True)
        +
        layout(f"""\

            Note - unused, in this case, would be ['Auckland', 'Mt Albert']

            """)
        )

def get_general_comprehension_msg():
    return layout("""\
        Comprehensions are one the great things about Python. To see why, have a
        look at Raymond Hettinger's classic talk "Transforming Code into
        Beautiful, Idiomatic Python" <https://youtu.be/OSGv2VnC0go?t=2738> where
        he explains the rationale. In short, if the goal of your code can be
        expressed as a single English sentence then it might belong on one line.
        The code should say what it is doing more than how it is doing it.
        Comprehensions are declarative and that is A Very Good Thing™.

        Pro tip: don't make comprehensions *in*comprehensions ;-). If your
        comprehension is hard to read it is probably better rewritten as a
        looping structure.
        """)

def get_list_comprehension_msg():
    return (
        layout("""\

            #### Example List Comprehension:

            """)
        +
        layout("""\
            names_lengths = [
                len(name)
                for name in ['Tinky Winky', 'Dipsy', 'La La', 'Po']
            ]
            """, is_code=True)
        +
        layout("""\

            produces an ordinary list:

            """)
        +
        layout(str(
            {
                len(name)
                for name in ['Tinky Winky', 'Dipsy', 'La La', 'Po']
            }
            ))
        +
        layout("""\

            It is also possible to add a simple filter using `if`

            """)
        +
        layout("""\
            names_lengths = [
                len(name)
                for name in ['Tinky Winky', 'Dipsy', 'La La', 'Po']
                if not name.startswith('T')
            ]
            """, is_code=True)
        +
        layout("""\

            produces an ordinary list:

            """)
        +
        layout(str(
            {
                len(name)
                for name in ['Tinky Winky', 'Dipsy', 'La La', 'Po']
                if not name.startswith('T')
            }
            ))
    )

def get_dict_comprehension_msg():
    return (
        layout("""\

        #### Example Dictionary Comprehension:

        """)
        +
        layout("""\
            country2capital = {{
                country: capital
                for country, capital in [('NZ', 'Wellington'), ('Italy', 'Rome')]
            }}
            """, is_code=True)
        +
        layout("""\

            produces an ordinary dictionary:

            """)
        +
        layout(str(
            {
                country: capital
                for country, capital
                in [('NZ', 'Wellington'), ('Italy', 'Rome')]
            }
            ))
        +
        layout("""\

            It is also possible to add a simple filter using `if`

            """)
        +
        layout("""\
            country2capital = {{
                country: capital
                for country, capital in [('NZ', 'Wellington'), ('Italy', 'Rome')]
                if country == 'NZ'
            }}
            """, is_code=True)
        +
        layout("""\

            produces an ordinary dictionary:

            """)
        +
        layout(str(
            {
                country: capital
                for country, capital
                in [('NZ', 'Wellington'), ('Italy', 'Rome')]
                if country == 'NZ'
            }
            ))
    )

def get_set_comprehension_msg():
    return (
        layout("""\

            #### Example Set Comprehension

            """)
        +
        layout("""\
            pets = {{
                pet for _person, pet
                in [('Rachel', 'cat'), ('Elliot', 'goat'), ('Giles', 'cat'),]
            }}
            """, is_code=True)
        +
        layout("""\

            produces an ordinary set (i.e. unique members only):

            """)
        +
        layout(str(
            {
                pet for _person, pet
                    in [('Rachel', 'cat'), ('Elliot', 'goat'), ('Giles', 'cat'),]
            }
            ))
        +
        layout("""\

            It is also possible to add a simple filter using `if`

            """)
        +
        layout("""\
            pets = {{
                pet for person, pet
                in [('Rachel', 'cat'), ('Elliot', 'goat'), ('Giles', 'cat'),]
                if person != 'Elliot'
            }}
            """, is_code=True)
        +
        layout("""\

            produces an ordinary set (i.e. unique members only):

            """)
        +
        layout(str(
            {
                pet for person, pet
                    in [('Rachel', 'cat'), ('Elliot', 'goat'), ('Giles', 'cat'),]
                    if person != 'Elliot'
            }
            ))
    )



from ..helpers import filt_block_help
from .. import ast_funcs, code_execution, conf
from ..gen_utils import layout_comment as layout, get_nice_str_list, int2nice

def truncate_set(items):
    return set(list(items)[: conf.MAX_ITEMS_EVALUATED])

ASSIGN_FUNC_NAME_XPATH = 'descendant-or-self::Assign/value/Call/func/Name'

@filt_block_help(xpath=ASSIGN_FUNC_NAME_XPATH)
def set_overview(block_dets, *, repeat=False):
    """
    Look for sets and provide general advice on using them and finding out more.
    """
    func_type_els = block_dets.element.xpath(ASSIGN_FUNC_NAME_XPATH)
    set_els = [func_type_el for func_type_el in func_type_els
        if func_type_el.get('id') == 'set']
    names_items, oversized_msg = code_execution.get_collections_dets(
        set_els, block_dets,
        collection_plural='sets', truncated_items_func=truncate_set)
    name_sets = [
        (name, items) for name, items in names_items if items is not None]
    if not name_sets:
        return None

    title = layout("""\
    ### Set details
    """)
    summary_bits = []
    for name, my_set in name_sets:
        empty_set = len(my_set) == 0
        if empty_set:
            summary_bits.append(layout(f"""\
            `{name}` is an empty set.
            """))
        else:
            members = str(sorted(my_set)).strip('[').strip(']')
            summary_bits.append(layout(f"""\

            `{name}` is a set with {int2nice(len(my_set))} members: {members}
            """))
    summary = ''.join(summary_bits)
    if not repeat:
        sets_rock = (
            layout("""\

            Python sets are brilliant. There are often cases in programming
            where you need some sort of set operation e.g. you need everything
            in one set that is not in another. In Python you can express that
            idea directly and semantically with set concepts instead of having
            to build the operations yourself in code which needs explaining and
            testing.

            For example:
            """)
            +
            layout("""\
            people = set(['Sam', 'Avi', 'Terri', 'Noor', 'Hyeji'])
            no_email = set(['Sam', 'Terri'])
            people2email = people - no_email
            ## >>> {'Noor', 'Hyeji', 'Avi'}
            """, is_code=True)
        )
        no_duplicates_demo = ''
        for name, my_set in name_sets:
            empty_set = len(my_set) == 0
            if not empty_set:
                members = str(sorted(my_set)).strip('[').strip(']')
                existing_set_item = list(my_set)[0]
                no_duplicates_demo = (
                    layout("""\
                    For example:
                    """)
                    +
                    layout(f"""\
                    {name}.add({existing_set_item})
                    ## >>> {my_set}
                    """, is_code=True)
                )
                break
        set_dets = (
            layout(f"""\

            Being a set, all members are unique by definition so if you add
            something to a set that is already a member the set doesn't change.
            """)
            +
            no_duplicates_demo
            +
            layout("""\

            This is unlike a list which grows each time you append an item - the
            list simply repeats the item. For example:
            """)
            +
            layout("""\
            my_list = [1, 2, 3]
            my_list.append(4)
            my_list.append(4)
            my_list.append(4)
            ## >>> [1, 2, 3, 4, 4, 4]
            my_set = {1, 2, 3}
            my_set.add(4)
            my_set.add(4)
            my_set.add(4)
            ## >>> {1, 2, 3, 4}
            """, is_code=True)
        )
        set_extras = (
            layout("""\

            Set operations can be expressed with operators such as `-` (minus)
            or with methods such as .difference().

            Sets are well explained in the official documentation so it will
            suffice to show a simple example.

            Note - the operator corresponding to the `.union()` method is the
            pipe `|` not `+`.
            """)
            +
            layout("""\
            badminton_players = set(['Grant', 'Charlotte', 'Aravind'])
            tennis_players = set(['Giles', 'Grant'])
            squash_players = set(['Grzegorz'])
            racquet_players = badminton_players | tennis_players | squash_players
            ## >>> {'Grzegorz', 'Giles', 'Grant', 'Aravind', 'Charlotte'}
            """, is_code=True)
        )
    else:
        sets_rock = ''
        set_dets = ''
        set_extras = ''

    message = {
        conf.BRIEF: title + oversized_msg + summary + sets_rock,
        conf.MAIN: title + oversized_msg + summary + sets_rock + set_dets,
        conf.EXTRA: set_extras,
    }
    return message

## Raise exceptions when something happens that apparently shouldn't

def _checking_non_membership(compare_el):
    """
    See if checking membership status (specifically non-membership status).
    """
    ops_els = compare_el.xpath('ops')
    if len(ops_els) != 1:
        raise Exception("Should only be one ops item in a Compare")
    ops_el = ops_els[0]
    ops_type_els = ops_el.getchildren()
    if len(ops_type_els) != 1:
        raise Exception("Should only be one ops child")
    ops_type = ops_type_els[0].tag
    checking_non_membership = (ops_type == 'NotIn')
    return checking_non_membership

def _append_after_check(call_el):
    func_attribute_els = call_el.xpath('func/Attribute')
    if len(func_attribute_els) != 1:
        return False
    func_attribute_el = func_attribute_els[0]
    if func_attribute_el.get('attr') != 'append':
        return False
    return True

def _get_test_item_dict(compare_el):
    left_els = compare_el.xpath('left')
    if len(left_els) != 1:
        raise Exception("Should only be one left item in a Compare")
    left_el = left_els[0]
    left_val_els = left_el.getchildren()
    if len(left_val_els) != 1:
        raise Exception("Should only be one value item in a Compare/left")
    left_val_el = left_val_els[0]
    test_item_dict = ast_funcs.get_standardised_el_dict(left_val_el)
    return test_item_dict

def _get_append_item_dict(call_el):
    args_els = call_el.xpath('args')
    if len(args_els) != 1:
        return None
    args_el = args_els[0]
    args_children_els = args_el.getchildren()
    if len(args_children_els) != 1:
        return None
    args_child_el = args_children_els[0]
    append_item_dict = ast_funcs.get_standardised_el_dict(args_child_el)
    return append_item_dict

def _get_test_collection_dict(compare_el):
    comparator_name_els = compare_el.xpath('comparators/Name')
    if len(comparator_name_els) != 1:
        return None
    comparator_name_el = comparator_name_els[0]
    test_collection_dict = ast_funcs.get_standardised_el_dict(
        comparator_name_el)
    return test_collection_dict

def _get_append_collection_dict(call_el):
    appended_name_els = call_el.xpath('func/Attribute/value/Name')
    if len(appended_name_els) != 1:
        return None
    appended_name_el = appended_name_els[0]
    append_collection_dict = ast_funcs.get_standardised_el_dict(
        appended_name_el)
    return append_collection_dict

def _get_inappropriate_list(if_el):
    """
    Look for a test item being identified as not being in a collection and then
    being appended to that collection.
    """
    test_el = if_el.xpath('test')[0]

    compare_els = test_el.xpath('Compare')
    if len(compare_els) != 1:
        return None
    compare_el = compare_els[0]

    checking_non_membership = _checking_non_membership(compare_el)
    if not checking_non_membership:
        return None

    body_el = if_el.xpath('body')[0]
    call_els = body_el.xpath('Expr/value/Call')
    if len(call_els) != 1:
        return None
    call_el = call_els[0]

    if not _append_after_check(call_el):
        return None

    test_item_dict = _get_test_item_dict(compare_el)
    append_item_dict = _get_append_item_dict(call_el)
    if append_item_dict is None:
        return None
    same_item_checked_and_appended = (test_item_dict == append_item_dict)
    if not same_item_checked_and_appended:
        return None

    test_collection_dict = _get_test_collection_dict(compare_el)
    if test_collection_dict is None:
        return None
    append_collection_dict = _get_append_collection_dict(call_el)
    if append_collection_dict is None:
        return None
    same_collection_in_membership_test_and_append = (
        test_collection_dict == append_collection_dict)
    if not same_collection_in_membership_test_and_append:
        return None

    inappropriate_list = test_collection_dict.get('id')
    return inappropriate_list

XPATH_COMPARE = 'descendant-or-self::If/test/Compare'

@filt_block_help(xpath=XPATH_COMPARE, warning=True)
def set_better_than_list(block_dets, *, repeat=False):
    """
    Look for cases where the code checks list membership before adding.
    Candidate for a set?
    """
    if_els = block_dets.element.xpath('descendant-or-self::If')
    inappropriate_lists = []
    for if_el in if_els:
        inappropriate_list = _get_inappropriate_list(if_el)
        if inappropriate_list is not None:
            inappropriate_lists.append(inappropriate_list)
    if not inappropriate_lists:
        return None

    title = layout("""\
    ### Using a `set` with the `add` method probably a better option
    """)
    first_list = inappropriate_lists[0]
    multiple = len(inappropriate_lists) > 1
    if multiple:
        dubious_lists_msg = get_nice_str_list(inappropriate_lists, quoter='`')
        summary = layout(f"""\

        It looks like the following collections are built by checking potential
        new values for non-membership and appending to them if not already
        present: {dubious_lists_msg}. Using a `set` could result in a more
        semantic option.
        """)
    else:
        summary = layout(f"""\

        It looks like `{first_list}` is built by checking potential new values
        for non-membership and appending if not already present. Using a `set`
        could result in a more semantic option.
        """)
    if not repeat:
        alternative = (layout("""\
            For example, the following:
            """)
            +
            layout(f"""\
            {first_list} = [value, ...]
            if value not in {first_list}:
                {first_list}.append(value)
            """, is_code=True)
            +
            layout("""\
            could be replaced with the simpler and more semantic:
            """)
            +
            layout(f"""\
            {first_list} = {{value, ...}}
            {first_list}.add(value)
            """, is_code=True)
            + layout("""\

            Adding is a no-op if a value is already a member of a `set`.

            The curly braces indicate a `set` if it contains individual items
            and a `dict` if it contains key: value pairs.
            """)
        )
    else:
        alternative = ''

    message = {
        conf.BRIEF: title + summary,
        conf.MAIN: title + summary + alternative,
    }
    return message



from collections import defaultdict, namedtuple

from superhelp.helpers import all_blocks_help, any_block_help, is_reserved_name
from .. import ast_funcs, conf, name_utils
from superhelp import gen_utils
from superhelp.gen_utils import (get_nice_str_list, int2first_etc, int2nice,
    layout_comment as layout)

PairDets = namedtuple('PairDets', 'name, name_value, unpacking_idx')
PairDets.name.__doc__ = (
    "Name assigned to the other name e.g. pet when pet = cat")
PairDets.name_value.__doc__ = ("Name acting as the value having the other name "
    "assigned to it e.g. cat when pet = cat")
PairDets.unpacking_idx.__doc__ = ("The unpacking index if applicable "
    "e.g if a, b = c and name is a then unpacking_idx is 0; for b it is 1")

ASSIGN_SUBSCRIPT_XPATH = 'descendant-or-self::Assign/value/Subscript'
ASSIGN_NAME_XPATH = 'descendant-or-self::Assign/value/Name'

def get_subscript_name_value(assign_subscript_el):
    """
    Looking for:
    people[0] (list) as in `name = people[0]`
    capitals['NZ'] (dict) as in `name = capitals['NZ']`
    """
    name_value_el = assign_subscript_el.xpath('value/Name')[0]
    name_value_name = name_value_el.get('id')
    try:
        slice_dets = ast_funcs.get_slice_dets(assign_subscript_el)
    except Exception:
        return None
    name_value = f"{name_value_name}{slice_dets}"
    return name_value

def pairs_dets_from_el(name2name_el):
    """
    Get details of a name to name assignment. Note - might be multiple names
    assigned to the name_value e.g. a, b = var

    :return: list of tuples of name, name_value. name_value might be a standard
     Python name e.g. person, or a subscript of some sort e.g. people[0] or
     capitals['NZ']
    """
    ancestor_assign_els = name2name_el.xpath('ancestor::Assign')
    try:
        ancestor_assign_el = ancestor_assign_els[-1]
    except IndexError:
        raise IndexError(  ## in theory, guaranteed to be one given how we got the name2name_els
            "Unable to identify ancestor Assign for name-to-name assignment")
    try:
        names_dets = name_utils.get_assigned_names(name2name_el)
    except Exception as e:
        raise Exception("Unable to identify name for name-value assignment. "
            f"Orig error: {e}")

    ## If there is a subscript there might have been assignment to a dict or list
    assign_subscript_els = ancestor_assign_el.xpath(ASSIGN_SUBSCRIPT_XPATH)
    if assign_subscript_els:
        assign_subscript_el = assign_subscript_els[0]
        name_value = get_subscript_name_value(assign_subscript_el)
    else:
        assign_name_els = ancestor_assign_el.xpath(ASSIGN_NAME_XPATH)
        direct_assign_el = assign_name_els[0]
        name_value = direct_assign_el.get('id')
    if not name_value:
        raise Exception(
            "Unable to identify name_value even though there should be one")
    pairs_dets = [
        PairDets(name_dets.name_str, name_value, name_dets.unpacking_idx)
            for name_dets in names_dets]
    return pairs_dets

def pairs_dets_from_block(block_el):
    """
    Get details of name-to-name assignments.

    Examples of name to name assignment:

    animal = pet
    first_person = people[0]
    capital_nz = capitals['NZ']

    Obviously, being Python, the assignment is to the value under the name but
    it is a name nonetheless.
    """
    name2name_els = block_el.xpath(
        f"{ASSIGN_SUBSCRIPT_XPATH}|{ASSIGN_NAME_XPATH}")
    if not name2name_els:
        return None
    pairs_dets = []
    for el in name2name_els:
        try:
            el_pairs_dets = pairs_dets_from_el(el)
        except Exception:
            continue
        else:
            pairs_dets.extend(el_pairs_dets)
    return pairs_dets

@all_blocks_help()
def names_and_values(blocks_dets):
    """
    Look for names assigned to other names and explain names and values in
    Python.
    """
    name2name_pairs_dets = []
    for block_dets in blocks_dets:
        block_el = block_dets.element
        block_name2name_pairs_dets = pairs_dets_from_block(block_el)
        if block_name2name_pairs_dets:
            name2name_pairs_dets.extend(block_name2name_pairs_dets)
    if not name2name_pairs_dets:
        return None

    name2name_pair_strs = []
    for pair_dets in name2name_pairs_dets:
        if pair_dets.unpacking_idx is not None:
            first_etc_str = int2first_etc(pair_dets.unpacking_idx + 1)
            name2name_pair_str = (
                f"`{pair_dets.name}` is assigned to the {first_etc_str} item of"
                f" `{pair_dets.name_value}` through value unpacking")
        else:
            name2name_pair_str = (
                f"`{pair_dets.name}` is assigned to `{pair_dets.name_value}`")
        name2name_pair_strs.append(name2name_pair_str)
    first_pair_dets = name2name_pairs_dets[0]
    if first_pair_dets.unpacking_idx is None:
        example = layout(f"""\

        So when your code assigned name `{first_pair_dets.name}` to another name
        `{first_pair_dets.name_value}`, for example, it was actually linking
        `{first_pair_dets.name}` to the **value** `{first_pair_dets.name_value}`
        was linked to. If you reassigned `{first_pair_dets.name_value}` to
        something else it would not affect `{first_pair_dets.name}`.
        """)
    else:
        first_etc_str = int2first_etc(first_pair_dets.unpacking_idx + 1)
        example = layout(f"""\

        So when your code assigned name `{first_pair_dets.name}` to the
        {first_etc_str} item in another name `{first_pair_dets.name_value}`, for
        example, it was actually linking `{first_pair_dets.name}` to the
        **value** of that item. If you replaced that item with something else it
        would not affect `{first_pair_dets.name}`.
        """)
    nice_name2name_pairs = get_nice_str_list(
        name2name_pair_strs, item_glue='; ', quoter='')

    title = layout("""
    ### Names assigned to names
    """)
    assignments = layout(f"""\

    Your code assigned names to other names as values: {nice_name2name_pairs}.
    """)
    brief_summary = layout("""\

    To get a better understanding of what that means in Python see / read
    <https://nedbatchelder.com/text/names1.html>. Or look at this advice at a
    higher level of detail e.g. 'Main'.
    """)
    main_summary = (
        layout(f"""\

        So what does that mean in Python? Having a firm grasp of how names and
        values work in Python means you can confidently reason about your code
        and pre-emptively avoid numerous bugs.

        #### How variables work in Python - names and values

        Assignment is linking names and values e.g. `fname = "Maggie"`

        Another way of expressing the same idea:

        > "Assignment: make this name refer to that value" Ned Batchelder

        Names are something you can assign to a value.

        Multiple names can refer to the same value. No name is the "real" name
        in Python - they are all just labels to a value.

        Names can take a variety of forms. For example:
        """)
        +
        layout("""\
        fname = value
        mydict['fruit'] = value
        family.pet = value
        """, is_code=True)
        +
        layout("""
        are all names.

        A wide variety of data structures can be values to which names are
        assigned. For example:
        """)
        +
        layout("""\
        name = 1066
        name = 'cheerful'
        name = [1, 2, 3, 4]
        name = {'width': 12.5, 'height': 23.75}

        etc
        """, is_code=True)
        +
        layout(f"""

        The key idea is that we are always linking names to values - NEVER names
        to other names. Python has references (names pointing to objects) but
        not pointers (names pointing to other names). When we assign one name to
        another we are actually linking to the underlying value of the second
        name.

        And be very clear - assignment never **copies** data. It just links
        names to values. There are no exceptions to this rule.

        So if `a = 'cat'` and `b = a` then both `a` and `b` are labels assigned
        to the string 'cat'. If we change what label `a` is attached to we don't
        affect `b`. It remains linked to 'cat'.

        {example}

        The names aren't necessarily independent though. If a mutable object
        like a list has multiple labels then any changes to the value, i.e. the
        list, are shared by all labels because they are all linked to the same
        value.

        For example:
        """)
        +
        layout("""\
        fruit = ['apple', 'banana']
        ingredients = fruit
        fruit.append('cherry')
        ## What is ingredients now? Try this code and find out. If you
        ## understand the ideas above you should be able to reason it out with
        ## confidence.
        """, is_code=True)
    )
    ned_talk = layout("""\
    The talk to watch / read on the topic is:
    <https://nedbatchelder.com/text/names1.html>

    <https://lerner.co.il/2019/06/18/understanding-python-assignment/> is also
    helpful.

    A note about using the verb 'assign'? I recommend saying that we assign
    names to values rather than the other way around because it better fits how
    Python actually works. The central reality in Python is the value not the
    name. The name is nothing substantial - it's not a container for a value
    or anything like that - it is merely a label.
    """)

    message = {
        conf.BRIEF: title + assignments + brief_summary,
        conf.MAIN: title + assignments + main_summary,
        conf.EXTRA: ned_talk,
    }
    return message

def _get_shamed_names_title(reserved_names, bad_names, dubious_names):
    if not (reserved_names or bad_names or dubious_names):
        raise Exception(f"No shamed names supplied to {__name__}")
    n_reserved_names = len(reserved_names)
    n_bad_names = len(bad_names)
    n_dubious_names = len(dubious_names)
    if reserved_names:
        if n_reserved_names == 1:
            title = "Bad naming - use of reserved name."
        else:
            title = "Bad naming - use of reserved names."
    else:
        title = ''
    if title:
        if bad_names:
            if n_bad_names == 1:
                if n_dubious_names < 1:
                    title += ' Also un-pythonic name'
                else:
                    title += ' Also un-pythonic name(s)'
            else:
                title += ' Also un-pythonic names'
        elif dubious_names:
            if n_dubious_names == 1:
                title += ' Also a possibly un-pythonic name'
            else:
                title += ' Also other possibly un-pythonic names'
    else:
        if bad_names:
            if n_bad_names == 1:
                if n_dubious_names < 1:
                    title = 'Un-pythonic name'
                else:
                    title = 'Un-pythonic name(s)'
            else:
                title = 'Un-pythonic names'
        elif dubious_names:
            if n_dubious_names == 1:
                title = 'Possibly an un-pythonic name'
            else:
                title = 'Possibly some un-pythonic names'
    return title

def get_standard_assigned_names(block_dets):
    """
    Only get names where we expect standard pythonic naming. So not named tuple
    or class names, for example. We have to explicitly ignore named tuple names.
    Classes are automatically excluded because I haven't explicitly included
    them. They are stored differently e.g. <ClassDef ...name="ActuallyGoodName">
    """
    assigned_name_els = block_dets.element.xpath(
        'descendant-or-self::targets/Name | descendant-or-self::target/Name')
    assigned_names = []
    for name_el in assigned_name_els:
        ## exclude if named tuple - they are allowed "un-Pythonic" names
        try:
            assign_el = name_el.xpath('ancestor::Assign')[0]
        except IndexError:
            ## cannot be a named tuple
            pass
        else:
            func_name_els = assign_el.xpath('value/Call/func/Name')
            if func_name_els:
                func_names = [
                    func_name_el.get('id') for func_name_el in func_name_els]
                if 'namedtuple' in func_names:
                    continue
        ## not a named tuple 
        name = name_el.get('id')
        assigned_names.append(name)
    return assigned_names

def get_class_names(block_dets):
    class_els = block_dets.element.xpath('ClassDef')
    class_names = [class_el.get('name') for class_el in class_els]
    return class_names

def get_named_tuple_names(block_dets):
    assigned_name_els = block_dets.element.xpath(
        'descendant-or-self::Assign/targets/Name')
    named_tuple_names = []
    for name_el in assigned_name_els:
        assign_el = name_el.xpath('ancestor::Assign')[0]
        func_name_els = assign_el.xpath('value/Call/func/Name')
        if func_name_els:
            func_names = [
                func_name_el.get('id') for func_name_el in func_name_els]
            if 'namedtuple' not in func_names:
                continue
        name = name_el.get('id')
        named_tuple_names.append(name)
    return named_tuple_names

def get_unpacked_names(block_dets):
    unpacked_name_els = block_dets.element.xpath(
        'descendant-or-self::targets/Tuple/elts/Name'
        ' | '
        'descendant-or-self::target/Tuple/elts/Name')
    unpacked_names = [unpacked_name_el.get('id')
        for unpacked_name_el in unpacked_name_els]
    return unpacked_names

def get_def_func_names(block_dets):
    def_func_elements = block_dets.element.xpath(
        'descendant-or-self::FunctionDef')
    def_func_names = [name_el.get('name') for name_el in def_func_elements]
    return def_func_names

def get_all_names(block_dets, *, include_non_standard=False):
    """
    :param bool include_non_standard: if True include variables that aren't
     expected to follow standard Python naming conventions e.g. class or named
     tuple names.
    """
    assigned_names = get_standard_assigned_names(block_dets)
    unpacked_names = get_unpacked_names(block_dets)
    def_func_names = get_def_func_names(block_dets)
    all_names = assigned_names + unpacked_names + def_func_names
    if include_non_standard:
        class_names = get_class_names(block_dets)
        named_tuple_names = get_named_tuple_names(block_dets)
        all_names = all_names + class_names + named_tuple_names
    return all_names

@any_block_help(warning=True)
def unpythonic_name_check(block_dets, *, repeat=False):
    """
    Check names used for use of reserved words and camel case.
    """
    names = get_all_names(block_dets, include_non_standard=False)
    reserved_names = set()
    bad_names = set()
    dubious_names = set()
    for name in names:
        if is_reserved_name(name):
            reserved_names.add(name)
        all_lower_case = (name.lower() == name)
        all_upper_case = (name.upper() == name)
        bad_name = not (all_lower_case or all_upper_case)
        if bad_name:
            bad_names.add(name)
        elif all_upper_case:
            dubious_names.add(name)
    if not (reserved_names or bad_names or dubious_names):
        return None

    shamed_names_title = _get_shamed_names_title(
        reserved_names, bad_names, dubious_names)
    title = layout(f"""\
    ### {shamed_names_title}
    """)
    if reserved_names:
        reserved_names_listed = gen_utils.get_nice_str_list(
            sorted(reserved_names), quoter='`')
        reserved_comment = layout(f"""\
        Reserved name(s): {reserved_names_listed}
        """)
    else:
        reserved_comment = ''
    if bad_names:
        bad_names_listed = gen_utils.get_nice_str_list(
            sorted(bad_names), quoter='`')
        bad_comment = layout(f"""\
        Un-pythonic name(s): {bad_names_listed}
        """)
    else:
        bad_comment = ''
    if dubious_names:
        dubious_names_listed = gen_utils.get_nice_str_list(
            sorted(dubious_names), quoter='`')
        dubious_comment = layout(f"""\
        Possibly un-pythonic name(s): {dubious_names_listed}
        """)
    else:
        dubious_comment = ''
    if not repeat:
        snake_case = layout("""\

        Python variables should not named using reserved words e.g.
        `collections` or `sorted`.

        Generally speaking Python variables should be snake case - that is lower
        case, with multiple words joined by underscores e.g. `high_scores` (not
        `highScores` or `HighScores`)
        """)
        pascal = layout("""\
        In Python class names and named tuples are expected to be in Pascal Case
        (also known as upper camel case) rather than the usual snake case. E.g.
        `collections.ChainMap`

        Exceptions can also be made when a higher priority is consistency with
        other code e.g. a library the Python is ported from, or the non-Python
        code that Python is wrapping.
        """)
    else:
        snake_case = ''
        pascal = ''

    message = {
        conf.BRIEF: (title + reserved_comment + bad_comment
            + dubious_comment + snake_case + pascal),
        conf.MAIN: (title + reserved_comment + bad_comment
            + dubious_comment + snake_case + pascal),
    }
    return message

@any_block_help(warning=True)
def short_name_check(block_dets, *, repeat=False):
    """
    Check for short variable names.
    """
    names = get_all_names(block_dets, include_non_standard=True)
    short_names = defaultdict(set)
    for name in names:
        if len(name) < conf.MIN_BRIEF_NAME:
            short_names[len(name)].add(name)
    if not short_names:
        return None

    title = layout("""
    ### Short variable name
    """)
    short_comment_bits = []
    for length, names in sorted(short_names.items()):
        freq = len(names)
        multiple = (freq > 1)
        plural = 's' if length > 1 else ''
        if multiple:
            nice_list = get_nice_str_list(list(names), quoter='`')
            short_comment_bits.append((
                f"{int2nice(freq)} variables have names "
                f"{int2nice(length)} character{plural} long: {nice_list}."))
        else:
            name = names.pop()
            short_comment_bits.append(f"`{name}` is short (only "
                f"{int2nice(length)} character{plural} long).")
    short_comment = '; '.join(short_comment_bits)
    sometimes_ok = layout(f"""\

    Sometimes, short variable names are appropriate - even conventional - but
    they should be avoided outside of a few special cases. In your code:

    {short_comment}
    """)
    if not repeat:
        idiomatic = (
            layout("""\

            In many programming languages it is idiomatic to use `i`, `j`, and
            even `k` as increment counters. It is best to reserve `i` for
            incrementing starting at 0 and prefer `n` when starting at 1. But
            longer names should be used when they aid readability (i.e. usually
            ;-)).

            `k` and `v` are idiomatic in Python when iterating through
            dictionary items e.g.
            """)
            +
            layout("""\
                for k, v in my_dict.items():
                    ...
            """, is_code=True)
            +
            layout("""\

            But even they should probably be replaced with something more
            descriptive.

            And there are other cases specific to certain contexts e.g. using r
            for the result of a request e.g.
            """)
            +
            layout("""\
            r = requests.get('https://pypi.org/project/superhelp/')
            """, is_code=True)
            +
            layout("""\

            The main goal is readability, readability, readability. That is what
            should drive variable naming above all else. Only a modest weight
            should be given to speed of typing in any code that is going to be
            used or worked on in the future as opposed to quick exploratory code
            in a terminal / notebook etc.

            In the words of the legendary Donald Knuth:

            > "Programs are meant to be read by humans and only incidentally for
            computers to execute."

            Note - excessively long variable names can be unreadable as well.
            They may also indicate the code needs to be reworked.
            """)
        )
    else:
        idiomatic = ''

    message = {
        conf.BRIEF: title + sometimes_ok,
        conf.MAIN: title + sometimes_ok + idiomatic,
    }
    return message



from textwrap import dedent, indent
import webbrowser

from .. import conf, gen_utils

from markdown import markdown  ## https://coderbook.com/@marcus/how-to-render-markdown-syntax-as-html-using-python/ @UnresolvedImport

DETAIL_LEVEL2CLASS = {
    detail_level: f"help help-{detail_level}"
    for detail_level in conf.DETAIL_LEVELS}

BROWSER_HTML_WRAPPER = """\
<!DOCTYPE html>
<html lang="en">
{head}
<body>
{logo_svg}
<h1>SuperHELP - Help for Humans!</h1>
{intro}
<p>{missing_advice_message}</p>
<p>Toggle between different levels of detail.</p>
{radio_buttons}
{body_inner}
{visibility_script}
</body>
</html>"""

NOTEBOOK_HTML_WRAPPER = """\
<!DOCTYPE html>
<html lang="en">
{head}
<body>
<h1>Look here for some help on the snippet in the cell above</h1>
{intro}
<p>{missing_advice_message}</p>
<p>Toggle between different levels of detail.</p>
{radio_buttons}
{body_inner}
{visibility_script}
</body>
</html>"""

CODE_CSS = """\
/*From https://richleland.github.io/pygments-css/ */
.codehilite .hll { background-color: #ffffcc }
.codehilite {
  background: #f8f8f8;
  padding: 6px 4px 4px 12px;
}
.codehilite .c { color: #408080; font-style: italic } /* Comment */
.codehilite .err { border: 1px solid #FF0000 } /* Error */
.codehilite .k { color: #008000; font-weight: bold } /* Keyword */
.codehilite .o { color: #666666 } /* Operator */
.codehilite .ch { color: #408080; font-style: italic } /* Comment.Hashbang */
.codehilite .cm { color: #408080; font-style: italic } /* Comment.Multiline */
.codehilite .cp { color: #BC7A00 } /* Comment.Preproc */
.codehilite .cpf { color: #408080; font-style: italic } /* Comment.PreprocFile */
.codehilite .c1 { color: #408080; font-style: italic } /* Comment.Single */
.codehilite .cs { color: #408080; font-style: italic } /* Comment.Special */
.codehilite .gd { color: #A00000 } /* Generic.Deleted */
.codehilite .ge { font-style: italic } /* Generic.Emph */
.codehilite .gr { color: #FF0000 } /* Generic.Error */
.codehilite .gh { color: #000080; font-weight: bold } /* Generic.Heading */
.codehilite .gi { color: #00A000 } /* Generic.Inserted */
.codehilite .go { color: #888888 } /* Generic.Output */
.codehilite .gp { color: #000080; font-weight: bold } /* Generic.Prompt */
.codehilite .gs { font-weight: bold } /* Generic.Strong */
.codehilite .gu { color: #800080; font-weight: bold } /* Generic.Subheading */
.codehilite .gt { color: #0044DD } /* Generic.Traceback */
.codehilite .kc { color: #008000; font-weight: bold } /* Keyword.Constant */
.codehilite .kd { color: #008000; font-weight: bold } /* Keyword.Declaration */
.codehilite .kn { color: #008000; font-weight: bold } /* Keyword.Namespace */
.codehilite .kp { color: #008000 } /* Keyword.Pseudo */
.codehilite .kr { color: #008000; font-weight: bold } /* Keyword.Reserved */
.codehilite .kt { color: #B00040 } /* Keyword.Type */
.codehilite .m { color: #666666 } /* Literal.Number */
.codehilite .s { color: #BA2121 } /* Literal.String */
.codehilite .na { color: #7D9029 } /* Name.Attribute */
.codehilite .nb { color: #008000 } /* Name.Builtin */
.codehilite .nc { color: #0000FF; font-weight: bold } /* Name.Class */
.codehilite .no { color: #880000 } /* Name.Constant */
.codehilite .nd { color: #AA22FF } /* Name.Decorator */
.codehilite .ni { color: #999999; font-weight: bold } /* Name.Entity */
.codehilite .ne { color: #D2413A; font-weight: bold } /* Name.Exception */
.codehilite .nf { color: #0000FF } /* Name.Function */
.codehilite .nl { color: #A0A000 } /* Name.Label */
.codehilite .nn { color: #0000FF; font-weight: bold } /* Name.Namespace */
.codehilite .nt { color: #008000; font-weight: bold } /* Name.Tag */
.codehilite .nv { color: #19177C } /* Name.Variable */
.codehilite .ow { color: #AA22FF; font-weight: bold } /* Operator.Word */
.codehilite .w { color: #bbbbbb } /* Text.Whitespace */
.codehilite .mb { color: #666666 } /* Literal.Number.Bin */
.codehilite .mf { color: #666666 } /* Literal.Number.Float */
.codehilite .mh { color: #666666 } /* Literal.Number.Hex */
.codehilite .mi { color: #666666 } /* Literal.Number.Integer */
.codehilite .mo { color: #666666 } /* Literal.Number.Oct */
.codehilite .sa { color: #BA2121 } /* Literal.String.Affix */
.codehilite .sb { color: #BA2121 } /* Literal.String.Backtick */
.codehilite .sc { color: #BA2121 } /* Literal.String.Char */
.codehilite .dl { color: #BA2121 } /* Literal.String.Delimiter */
.codehilite .sd { color: #BA2121; font-style: italic } /* Literal.String.Doc */
.codehilite .s2 { color: #BA2121 } /* Literal.String.Double */
.codehilite .se { color: #BB6622; font-weight: bold } /* Literal.String.Escape */
.codehilite .sh { color: #BA2121 } /* Literal.String.Heredoc */
.codehilite .si { color: #BB6688; font-weight: bold } /* Literal.String.Interpol */
.codehilite .sx { color: #008000 } /* Literal.String.Other */
.codehilite .sr { color: #BB6688 } /* Literal.String.Regex */
.codehilite .s1 { color: #BA2121 } /* Literal.String.Single */
.codehilite .ss { color: #19177C } /* Literal.String.Symbol */
.codehilite .bp { color: #008000 } /* Name.Builtin.Pseudo */
.codehilite .fm { color: #0000FF } /* Name.Function.Magic */
.codehilite .vc { color: #19177C } /* Name.Variable.Class */
.codehilite .vg { color: #19177C } /* Name.Variable.Global */
.codehilite .vi { color: #19177C } /* Name.Variable.Instance */
.codehilite .vm { color: #19177C } /* Name.Variable.Magic */
.codehilite .il { color: #666666 } /* Literal.Number.Integer.Long */
"""

VISIBILITY_SCRIPT = """\
<script>
 function updateVerbosity() {
   var verbositySelectors = {
     '%(brief)s': '.help-%(brief)s',
     '%(main)s': '.help-%(main)s',
     '%(extra)s': '.help-%(main)s, .help-%(extra)s'
   }
   // Get selected verbosity.
   var verbosity = document.querySelector('input[name="verbosity"]:checked').value;
   // Hide all helps.
   document.querySelectorAll('.help').forEach(function(helpElement) {
     helpElement.classList.remove('help-visible');
   });
   // Show only selected helps.
   document.querySelectorAll(verbositySelectors[verbosity]).forEach(function(helpElement) {
     helpElement.classList.add('help-visible');
   });
 }

 // Update verbosity after page load.
 updateVerbosity();

 // Update verbosity whenever a radio is changed.
 var radios = document.querySelectorAll('input[name="verbosity"]');
 radios.forEach(function(radio) {
   radio.addEventListener('change', updateVerbosity);
 });
</script>
""" % {'brief': conf.BRIEF, 'main': conf.MAIN, 'extra': conf.EXTRA}

PART = 'part'
IS_CODE = 'is_code'

def _get_radio_buttons(*, detail_level=conf.BRIEF):
    radio_buttons_dets = []
    for message_type in conf.DETAIL_LEVELS:
        checked = ' checked' if message_type == detail_level else ''
        radio_button_dets = f"""\
            <input type="radio"
             id="radio-verbosity-{message_type}"
             name="verbosity"
             value="{message_type}"{checked}>
            <label for="radio-verbosity-{message_type}">{message_type}</label>
            """
        radio_buttons_dets.append(radio_button_dets)
    radio_buttons = '\n'.join(radio_buttons_dets)
    return radio_buttons

def get_separate_code_message_parts(message):
    """
    Need to handle code portions differently so need to separate it out.

    :return: list of message part details
     [{'part': 'asdf', 'is_code': False}, ...]
    :rtype: list
    """
    message_parts = []
    open_code_block = False
    lines = message.split('\n')
    for i, line in enumerate(lines):
        first_line = i == 0
        line_in_code_block = line.startswith(' ' * 4)
        if line_in_code_block:
            if open_code_block:
                open_code_part = message_parts[-1]
                open_code_part[PART] += f"\n{line}"  ## add to already-open code part
            else:
                message_parts.append({PART: line, IS_CODE: True})  ## start new code part
            open_code_block = True
        else:  ## line not in a code block
            if first_line or open_code_block:
                message_parts.append({PART: line, IS_CODE: False})
            else:
                open_non_code_part = message_parts[-1]
                open_non_code_part[PART] += f"\n{line}"
            open_code_block = False
    return message_parts

def get_html_strs(message, message_type, *, warning=False):  # @UnusedVariable
    if not message:
        return []
    message_type_class = DETAIL_LEVEL2CLASS[message_type]
    str_html_list = [f"<div class='{message_type_class}'>", ]
    message_parts = get_separate_code_message_parts(message)
    for message_part in message_parts:
        if message_part[IS_CODE]:
            message_part_str = markdown(
                message_part[PART], extensions=['codehilite'])
        else:
            message_part_str = markdown(
                dedent(message_part[PART]))
        str_html_list.append(message_part_str)
    str_html_list.append("</div>")
    return str_html_list

def get_message_html_strs(message_dets):
    """
    Process message.
    """
    message_html_strs = []
    if message_dets.warning:
        message_html_strs.append("<div class='warning'>")
    for detail_level in conf.DETAIL_LEVELS:
        try:
            message = message_dets.message[detail_level]
        except KeyError:
            if detail_level != conf.EXTRA:
                raise Exception(
                    f"Missing required message level {detail_level}")
        except TypeError:
            raise TypeError(
                f"Missing message in message_dets {message_dets}")
        else:
            try:
                message = (
                    message
                    .replace(conf.PYTHON_CODE_START, conf.MD_PYTHON_CODE_START)
                    .replace(f"\n    {conf.PYTHON_CODE_END}", '')
                )
            except Exception:
                pass
            detail_level_html_strs = get_html_strs(
                message, detail_level, warning=message_dets.warning)
            message_html_strs.extend(detail_level_html_strs)
    if message_dets.warning:
        message_html_strs.append("</div>")
    return message_html_strs

def repeat_overall_snippet(snippet, file_path):
    html_strs = []
    code_desc = gen_utils.get_code_desc(file_path)
    html_strs.append(f"<h2>{code_desc}</h2>")
    line_numbered_snippet = gen_utils.get_line_numbered_snippet(snippet)
    overall_code_str = indent(
        f"{conf.MD_PYTHON_CODE_START}\n{line_numbered_snippet}",
        ' '*4)
    overall_code_str_highlighted = markdown(
        overall_code_str, extensions=['codehilite'])
    html_strs.append(overall_code_str_highlighted)
    return html_strs

def _need_snippet_displayed(overall_messages_dets, block_messages_dets, *,
        in_notebook=False, multi_block=False):
    """
    Don't need to see the code snippet displayed when it is already visible
    right next to it:
    * because in notebook (it is already in the cell straight above)
    * because there is only one block in snippet and there is a block message
      for it (which will display the block i.e. the entire snippet) UNLESS there
      is an overall message separating them
    Otherwise we need it displayed.
    """
    if in_notebook:
        return False
    mono_block_snippet = not multi_block
    if mono_block_snippet and block_messages_dets and not overall_messages_dets:
        return False
    return True

def _get_all_html_strs(snippet, file_path,
        overall_messages_dets, block_messages_dets, *,
        warnings_only=False, in_notebook=False, multi_block=False):
    """
    Display all message types - eventually will show brief and, if the user
    clicks to expand, main instead with the option of expanding to show Extra.

    Suppress overall snippet display if in notebook - it is right above already
    and repeating it is confusing and obscures the feedback.
    """
    all_html_strs = []

    ## any feedback on user options chosen
    if not in_notebook:
        if warnings_only:
            options_msg = f"<p>{conf.WARNINGS_ONLY_MSG}.</p>"
        else:
            options_msg = f"<p>{conf.ALL_HELP_SHOWING_MSG}.</p>"
        all_html_strs.append(options_msg)
        all_html_strs.append("<div id='star'>"
            "Please give us a star <span id='grow'>&starf;</span> on "
            "<a href='https://github.com/grantps/superhelp'>"
            "SuperHELP's GitHub page</a></br>so others can learn about us. "
            "Thanks!</div>")
    ## overall snippet display
    display_snippet = _need_snippet_displayed(
        overall_messages_dets, block_messages_dets,
        in_notebook=in_notebook, multi_block=multi_block)
    if display_snippet:
        overall_snippet_html_strs = repeat_overall_snippet(snippet, file_path)
        all_html_strs.extend(overall_snippet_html_strs)

    ## overall messages
    overall_messages_dets.sort(key=lambda nt: nt.warning)
    for message_dets in overall_messages_dets:
        message_html_strs = get_message_html_strs(message_dets)
        all_html_strs.extend(message_html_strs)

    ## block messages
    block_messages_dets.sort(key=lambda nt: (nt.first_line_no, nt.warning))  ## by code block then within blocks warnings last
    prev_line_no = None
    for message_dets in block_messages_dets:
        ## display code for line number (once ;-)) Each line might have one or more messages but it will always have the one code_str starting on that line
        line_no = message_dets.first_line_no
        new_block = (line_no != prev_line_no)
        if new_block:
            block_has_warning_header = False
            all_html_strs.append(
                f'<h2>Code block starting line {line_no:,}</h2>')
            block_code_str = indent(
                f"{conf.MD_PYTHON_CODE_START}\n{message_dets.code_str}",
                ' '*4)
            block_code_str_highlighted = markdown(
                block_code_str, extensions=['codehilite'])
            all_html_strs.append(block_code_str_highlighted)
            prev_line_no = line_no
        if message_dets.warning and not block_has_warning_header:
            all_html_strs.append("<h3>Questions / Warnings</h3>")
            all_html_strs.append("<p>There may be some issues with this code "
                "block you want to address.</p>")
            block_has_warning_header = True
        message_html_strs = get_message_html_strs(message_dets)
        all_html_strs.extend(message_html_strs)
    return all_html_strs

def _get_head(*, in_notebook=False):
    internal_css = conf.INTERNAL_CSS % {
        'code_css': CODE_CSS,
        'margin_css': '' if in_notebook else 'margin: 40px 70px 20px 70px;',
        'max_width_css': '' if in_notebook else 'max-width: 700px;'
    }
    head = conf.HTML_HEAD % {
        'internal_css': internal_css}
    return head

def display(snippet, file_path, messages_dets, *,
        detail_level=conf.BRIEF,
        in_notebook=False, warnings_only=False,
        multi_block=False, multi_script=False):
    """
    Show for overall snippet and then by code blocks as appropriate.

    :param bool in_notebook: if True, return HTML as string; else open browser
     and display
    """
    raw_intro = gen_utils.get_intro(file_path, multi_block=multi_block)
    intro = f"<p>{raw_intro}</p>" if raw_intro else ''
    radio_buttons = _get_radio_buttons(detail_level=detail_level)
    overall_messages_dets, block_messages_dets = messages_dets
    all_html_strs = _get_all_html_strs(snippet, file_path,
        overall_messages_dets, block_messages_dets, warnings_only=warnings_only,
        in_notebook=in_notebook, multi_block=multi_block)
    body_inner = '\n'.join(all_html_strs)
    head = _get_head(in_notebook=in_notebook)
    if in_notebook:
        html2write = NOTEBOOK_HTML_WRAPPER.format(
            head=head,
            radio_buttons=radio_buttons,
            intro=intro,
            missing_advice_message=conf.MISSING_ADVICE_MESSAGE,
            body_inner=body_inner,
            visibility_script=VISIBILITY_SCRIPT)
        deferred_display = html2write
        return deferred_display
    else:
        html2write = BROWSER_HTML_WRAPPER.format(
            head=head, logo_svg=conf.LOGO_SVG,
            radio_buttons=radio_buttons,
            intro=intro,
            missing_advice_message=conf.MISSING_ADVICE_MESSAGE,
            body_inner=body_inner,
            visibility_script=VISIBILITY_SCRIPT)
        if file_path:
            raw_file_name = gen_utils.clean_path_name(file_path)
            file_name = f'{raw_file_name}.html'
        else:
            file_name = 'superhelp_output.html'
        superhelp_tmpdir = gen_utils.get_superhelp_tmpdir(
            folder=conf.SUPERHELP_GEN_OUTPUT)
        with gen_utils.make_open_tmp_file(
                file_name, superhelp_tmpdir=superhelp_tmpdir,
                mode='w') as tmp_dets:
            _superhelp_tmpdir, tmp_fh, fpath = tmp_dets
            tmp_fh.write(html2write)
        if multi_script:
            pass  ## display is deferred
        else:
            url = fpath.as_uri()
            webbrowser.open_new_tab(url)



from collections import defaultdict

from superhelp.helpers import get_unpacking_msg, all_blocks_help, \
    filt_block_help
from .. import ast_funcs, conf
from superhelp import gen_utils
from superhelp.gen_utils import layout_comment as layout

ASSIGN_UNPACKING_XPATH = 'descendant-or-self::Assign/targets/Tuple'

@filt_block_help(xpath=ASSIGN_UNPACKING_XPATH)
def unpacking(block_dets, *, repeat=False):
    """
    Identify name unpacking e.g. x, y = coord
    """
    unpacked_els = block_dets.element.xpath(ASSIGN_UNPACKING_XPATH)

    title = layout("""\
    ### Name unpacking
    """)
    summary_bits = []
    for unpacked_el in unpacked_els:
        unpacked_names = [
            name_el.get('id') for name_el in unpacked_el.xpath('elts/Name')]
        if not unpacked_names:
            continue
        nice_str_list = gen_utils.get_nice_str_list(unpacked_names, quoter='`')
        summary_bits.append(layout(f"""\

        Your code uses unpacking to assign names {nice_str_list}
        """))
    summary = ''.join(summary_bits)
    if not repeat:
        unpacking_msg = get_unpacking_msg()
    else:
        unpacking_msg = ''

    message = {
        conf.BRIEF: title + summary,
        conf.EXTRA: unpacking_msg,
    }
    return message

@all_blocks_help()
def unpacking_opportunity(blocks_dets):
    """
    Look for opportunities to unpack values into multiple names instead of
    repeated and un-pythonic extraction using indexes.

    Signs of an unpacking opportunity - something is repeatedly sliced with
    different slice numbers e.g.

    x, y = coord
    vs
    x = coord[0]
    y = coord[1]
    """
    source_slices = defaultdict(set)
    for block_dets in blocks_dets:
        assign_els = block_dets.element.xpath('descendant-or-self::Assign')
        for assign_el in assign_els:
            try:
                slice_source = assign_el.xpath(
                    'value/Subscript/value/Name')[0].get('id')
                slice_n = ast_funcs.get_slice_n(assign_el)
            except (IndexError, TypeError):
                continue
            else:
                source_slices[slice_source].add(slice_n)
    sources2unpack = [source
        for source, slice_ns in source_slices.items()
        if len(slice_ns) > 1]
    if not sources2unpack:
        return None

    title = layout("""\
    ### Unpacking opportunity
    """)
    multiple_items = len(sources2unpack) > 1
    if multiple_items:
        nice_sources_list = gen_utils.get_nice_str_list(sources2unpack, quoter='`')
        unpackable = layout(f"""\

        {nice_sources_list} have multiple items extracted by indexing so might
        be suitable candidates for unpacking.
        """)
    else:
        unpackable = layout(f"""\

        Name (variable) `{sources2unpack[0]}` has multiple items extracted by
        indexing so might be a suitable candidate for unpacking.
        """)

    message = {
        conf.BRIEF: title + unpackable,
        conf.EXTRA: get_unpacking_msg(),
    }
    return message
